package gas.world.blocks.power;

import gas.entities.comp.*;
import mindustry.entities.*;
import gas.type.*;
import gas.world.blocks.logic.*;
import gas.content.*;
import mindustry.world.blocks.defense.turrets.*;
import gas.world.blocks.payloads.*;
import gas.world.meta.*;
import mindustry.annotations.Annotations.*;
import gas.world.blocks.units.*;
import mindustry.world.blocks.heat.*;
import java.util.*;
import mindustry.core.*;
import mindustry.world.blocks.legacy.*;
import mindustry.gen.*;
import mindustry.world.blocks.production.*;
import mindustry.world.draw.*;
import arc.math.*;
import mindustry.world.blocks.liquid.*;
import mindustry.world.meta.*;
import gas.world.blocks.heat.*;
import gas.world.blocks.defense.*;
import mindustry.world.blocks.distribution.*;
import gas.world.blocks.power.*;
import mindustry.world.*;
import gas.world.blocks.sandbox.*;
import mindustry.world.blocks.storage.*;
import gas.world.blocks.liquid.*;
import gas.entities.*;
import mindustry.world.blocks.campaign.*;
import gas.world.blocks.defense.turrets.*;
import gas.world.blocks.distribution.*;
import gas.world.*;
import mindustry.world.consumers.*;
import gas.world.blocks.gas.*;
import arc.math.geom.*;
import gas.world.blocks.campaign.*;
import mindustry.world.modules.*;
import gas.ui.*;
import mindustry.world.blocks.environment.*;
import mindustry.*;
import gas.world.consumers.*;
import mindustry.world.blocks.payloads.*;
import mindustry.world.blocks.*;
import gas.world.blocks.production.GasGenericCrafter.*;
import arc.graphics.g2d.*;
import mindustry.world.blocks.logic.*;
import gas.world.modules.*;
import gas.world.blocks.*;
import arc.graphics.*;
import gas.*;
import gas.io.*;
import gas.world.draw.*;
import gas.gen.*;
import gas.world.blocks.storage.*;
import mindustry.world.blocks.units.*;
import mindustry.input.*;
import mindustry.world.blocks.power.BeamNode.*;
import mindustry.graphics.*;
import gas.world.blocks.production.*;
import arc.struct.*;
import mindustry.world.blocks.defense.*;
import gas.entities.bullets.*;
import mindustry.world.blocks.power.*;
import mindustry.world.blocks.sandbox.*;
import static mindustry.Vars.*;

public class GasBeamNode extends GasPowerBlock {

    public int range = 5;

    @Load("power-beam")
    public TextureRegion laser;

    @Load("power-beam-end")
    public TextureRegion laserEnd;

    public Color laserColor1 = Color.white;

    public Color laserColor2 = Color.valueOf("ffd9c2");

    public float pulseScl = 7, pulseMag = 0.05f;

    public float laserWidth = 0.4f;

    public GasBeamNode(String name) {
        super(name);
        consumesPower = outputsPower = false;
        drawDisabled = false;
        envEnabled |= Env.space;
        allowDiagonal = false;
        underBullets = true;
        priority = TargetPriority.transport;
    }

    @Override
    public void setBars() {
        super.setBars();
        addBar("power", PowerNode.makePowerBalance());
        addBar("batteries", PowerNode.makeBatteryBalance());
    }

    @Override
    public void init() {
        super.init();
        updateClipRadius((range + 1) * tilesize);
    }

    @Override
    public void drawPlace(int x, int y, int rotation, boolean valid) {
        for (int i = 0; i < 4; i++) {
            int maxLen = range + size / 2;
            Building dest = null;
            var dir = Geometry.d4[i];
            int dx = dir.x, dy = dir.y;
            int offset = size / 2;
            for (int j = 1 + offset; j <= range + offset; j++) {
                var other = world.build(x + j * dir.x, y + j * dir.y);
                if (other != null && other.block.hasPower && other.team == Vars.player.team() && !(other.block instanceof PowerNode)) {
                    maxLen = j;
                    dest = other;
                    break;
                }
            }
            Drawf.dashLine(Pal.placing, x * tilesize + dx * (tilesize * size / 2f + 2), y * tilesize + dy * (tilesize * size / 2f + 2), x * tilesize + dx * (maxLen) * tilesize, y * tilesize + dy * (maxLen) * tilesize);
            if (dest != null) {
                Drawf.square(dest.x, dest.y, dest.block.size * tilesize / 2f + 2.5f, 0f);
            }
        }
    }

    @Override
    public void changePlacementPath(Seq<Point2> points, int rotation, boolean diagonal) {
        if (!diagonal) {
            Placement.calculateNodes(points, this, rotation, (point, other) -> Math.max(Math.abs(point.x - other.x), Math.abs(point.y - other.y)) <= range);
        }
    }

    public class GasBeamNodeBuild extends GasBuilding {

        // current links in cardinal directions
        public Building[] links = new Building[4];

        public Tile[] dests = new Tile[4];

        public int lastChange = -2;

        @Override
        public void updateTile() {
            // TODO this block technically does not need to update every frame, perhaps put it in a special list.
            if (lastChange != world.tileChanges) {
                lastChange = world.tileChanges;
                updateDirections();
            }
        }

        @Override
        public BlockStatus status() {
            if (Mathf.equal(power.status, 0f, 0.001f))
                return BlockStatus.noInput;
            if (Mathf.equal(power.status, 1f, 0.001f))
                return BlockStatus.active;
            return BlockStatus.noOutput;
        }

        @Override
        public void draw() {
            super.draw();
            if (Mathf.zero(Renderer.laserOpacity))
                return;
            Draw.z(Layer.power);
            Draw.color(laserColor1, laserColor2, (1f - power.graph.getSatisfaction()) * 0.86f + Mathf.absin(3f, 0.1f));
            Draw.alpha(Renderer.laserOpacity);
            float w = laserWidth + Mathf.absin(pulseScl, pulseMag);
            for (int i = 0; i < 4; i++) {
                if (dests[i] != null && (!(links[i].block instanceof GasBeamNode node) || (links[i].tileX() != tileX() && links[i].tileY() != tileY()) || (links[i].id > id && range >= node.range) || range > node.range)) {
                    int dst = Math.max(Math.abs(dests[i].x - tile.x), Math.abs(dests[i].y - tile.y));
                    // don't draw lasers for adjacent blocks
                    if (dst > 1 + size / 2) {
                        var point = Geometry.d4[i];
                        float poff = tilesize / 2f;
                        Drawf.laser(laser, laserEnd, x + poff * size * point.x, y + poff * size * point.y, dests[i].worldx() - poff * point.x, dests[i].worldy() - poff * point.y, w);
                    }
                }
            }
            Draw.reset();
        }

        @Override
        public void pickedUp() {
            Arrays.fill(links, null);
            Arrays.fill(dests, null);
        }

        public void updateDirections() {
            for (int i = 0; i < 4; i++) {
                var prev = links[i];
                var dir = Geometry.d4[i];
                links[i] = null;
                dests[i] = null;
                int offset = size / 2;
                // find first block with power in range
                for (int j = 1 + offset; j <= range + offset; j++) {
                    var other = world.build(tile.x + j * dir.x, tile.y + j * dir.y);
                    // power nodes do NOT play nice with beam nodes, do not touch them as that forcefully modifies their links
                    if (other != null && other.block.hasPower && other.team == team && !(other.block instanceof PowerNode)) {
                        links[i] = other;
                        dests[i] = world.tile(tile.x + j * dir.x, tile.y + j * dir.y);
                        break;
                    }
                }
                var next = links[i];
                if (next != prev) {
                    // unlinked, disconnect and reflow
                    if (prev != null) {
                        prev.power.links.removeValue(pos());
                        power.links.removeValue(prev.pos());
                        PowerGraph newgraph = new PowerGraph();
                        // reflow from this point, covering all tiles on this side
                        newgraph.reflow(this);
                        if (prev.power.graph != newgraph) {
                            // reflow power for other end
                            PowerGraph og = new PowerGraph();
                            og.reflow(prev);
                        }
                    }
                    // linked to a new one, connect graphs
                    if (next != null) {
                        power.links.addUnique(next.pos());
                        next.power.links.addUnique(pos());
                        power.graph.addGraph(next.power.graph);
                    }
                }
            }
        }
    }
}
