package gas.world.blocks.production;

import mindustry.annotations.Annotations.*;
import mindustry.entities.units.*;
import gas.entities.comp.*;
import gas.type.*;
import gas.world.blocks.logic.*;
import gas.content.*;
import mindustry.type.*;
import gas.world.blocks.payloads.*;
import arc.*;
import gas.world.meta.*;
import mindustry.ui.*;
import gas.world.blocks.units.*;
import mindustry.world.blocks.heat.*;
import arc.util.*;
import mindustry.world.blocks.legacy.*;
import mindustry.gen.*;
import mindustry.world.blocks.production.*;
import mindustry.world.draw.*;
import arc.math.*;
import mindustry.world.blocks.liquid.*;
import mindustry.world.meta.*;
import gas.world.blocks.heat.*;
import gas.world.blocks.defense.*;
import mindustry.world.blocks.distribution.*;
import gas.world.blocks.power.*;
import mindustry.world.*;
import gas.world.blocks.sandbox.*;
import mindustry.world.blocks.storage.*;
import gas.world.blocks.liquid.*;
import mindustry.game.*;
import gas.entities.*;
import mindustry.world.blocks.campaign.*;
import gas.world.blocks.defense.turrets.*;
import gas.world.blocks.distribution.*;
import gas.world.*;
import mindustry.world.consumers.*;
import mindustry.world.blocks.defense.turrets.*;
import gas.world.blocks.gas.*;
import arc.math.geom.*;
import gas.world.blocks.campaign.*;
import mindustry.world.modules.*;
import gas.ui.*;
import mindustry.world.blocks.environment.*;
import gas.world.consumers.*;
import mindustry.world.blocks.payloads.*;
import mindustry.world.blocks.*;
import gas.world.blocks.production.GasGenericCrafter.*;
import arc.graphics.g2d.*;
import mindustry.world.blocks.logic.*;
import gas.world.modules.*;
import gas.world.blocks.*;
import arc.graphics.*;
import gas.*;
import gas.io.*;
import gas.world.draw.*;
import mindustry.world.blocks.production.BeamDrill.*;
import gas.gen.*;
import gas.world.blocks.storage.*;
import mindustry.world.blocks.units.*;
import mindustry.graphics.*;
import gas.world.blocks.production.*;
import arc.util.io.*;
import mindustry.world.blocks.defense.*;
import gas.entities.bullets.*;
import mindustry.world.blocks.power.*;
import mindustry.world.blocks.sandbox.*;
import static mindustry.Vars.*;

public class GasBeamDrill extends GasBlock {

    protected Rand rand = new Rand();

    @Load("drill-laser")
    public TextureRegion laser;

    @Load("drill-laser-end")
    public TextureRegion laserEnd;

    @Load("drill-laser-center")
    public TextureRegion laserCenter;

    @Load("drill-laser-boost")
    public TextureRegion laserBoost;

    @Load("drill-laser-boost-end")
    public TextureRegion laserEndBoost;

    @Load("drill-laser-boost-center")
    public TextureRegion laserCenterBoost;

    @Load("@-top")
    public TextureRegion topRegion;

    @Load("@-glow")
    public TextureRegion glowRegion;

    public float drillTime = 200f;

    public int range = 5;

    public int tier = 1;

    public float laserWidth = 0.65f;

    /**
     * How many times faster the drill will progress when boosted by an optional consumer.
     */
    public float optionalBoostIntensity = 2.5f;

    public Color sparkColor = Color.valueOf("fd9e81"), glowColor = Color.white;

    public float glowIntensity = 0.2f, pulseIntensity = 0.07f;

    public float glowScl = 3f;

    public int sparks = 7;

    public float sparkRange = 10f, sparkLife = 27f, sparkRecurrence = 4f, sparkSpread = 45f, sparkSize = 3.5f;

    public Color boostHeatColor = Color.sky.cpy().mul(0.87f);

    public Color heatColor = new Color(1f, 0.35f, 0.35f, 0.9f);

    public float heatPulse = 0.3f, heatPulseScl = 7f;

    public GasBeamDrill(String name) {
        super(name);
        hasItems = true;
        rotate = true;
        update = true;
        solid = true;
        drawArrow = false;
        regionRotated1 = 1;
        envEnabled |= Env.space;
    }

    @Override
    public void init() {
        updateClipRadius((range + 2) * tilesize);
        super.init();
    }

    @Override
    public void setBars() {
        super.setBars();
        addBar("drillspeed", (GasBeamDrillBuild e) -> new Bar(() -> Core.bundle.format("bar.drillspeed", Strings.fixed(e.lastDrillSpeed * 60, 2)), () -> Pal.ammo, () -> e.warmup));
    }

    @Override
    public boolean outputsItems() {
        return true;
    }

    @Override
    public boolean rotatedOutput(int x, int y) {
        return false;
    }

    @Override
    public TextureRegion[] icons() {
        return new TextureRegion[] { region, topRegion };
    }

    @Override
    public void drawPlanRegion(BuildPlan plan, Eachable<BuildPlan> list) {
        Draw.rect(region, plan.drawx(), plan.drawy());
        Draw.rect(topRegion, plan.drawx(), plan.drawy(), plan.rotation * 90);
    }

    @Override
    public void setStats() {
        super.setStats();
        if (optionalBoostIntensity != 1) {
            stats.add(Stat.boostEffect, optionalBoostIntensity, StatUnit.timesSpeed);
        }
    }

    @Override
    public void drawPlace(int x, int y, int rotation, boolean valid) {
        Item item = null, invalidItem = null;
        boolean multiple = false;
        int count = 0;
        for (int i = 0; i < size; i++) {
            nearbySide(x, y, rotation, i, Tmp.p1);
            int j = 0;
            Item found = null;
            for (; j < range; j++) {
                int rx = Tmp.p1.x + Geometry.d4x(rotation) * j, ry = Tmp.p1.y + Geometry.d4y(rotation) * j;
                Tile other = world.tile(rx, ry);
                if (other != null && other.solid()) {
                    Item drop = other.wallDrop();
                    if (drop != null) {
                        if (drop.hardness <= tier) {
                            found = drop;
                            count++;
                        } else {
                            invalidItem = drop;
                        }
                    }
                    break;
                }
            }
            if (found != null) {
                // check if multiple items will be drilled
                if (item != found && item != null) {
                    multiple = true;
                }
                item = found;
            }
            int len = Math.min(j, range - 1);
            Drawf.dashLine(found == null ? Pal.remove : Pal.placing, Tmp.p1.x * tilesize, Tmp.p1.y * tilesize, (Tmp.p1.x + Geometry.d4x(rotation) * len) * tilesize, (Tmp.p1.y + Geometry.d4y(rotation) * len) * tilesize);
        }
        if (item != null) {
            float width = drawPlaceText(Core.bundle.formatFloat("bar.drillspeed", 60f / drillTime * count, 2), x, y, valid);
            if (!multiple) {
                float dx = x * tilesize + offset - width / 2f - 4f, dy = y * tilesize + offset + size * tilesize / 2f + 5, s = iconSmall / 4f;
                Draw.mixcol(Color.darkGray, 1f);
                Draw.rect(item.fullIcon, dx, dy - 1, s, s);
                Draw.reset();
                Draw.rect(item.fullIcon, dx, dy, s, s);
            }
        } else if (invalidItem != null) {
            drawPlaceText(Core.bundle.get("bar.drilltierreq"), x, y, false);
        }
    }

    @Override
    public boolean canPlaceOn(Tile tile, Team team, int rotation) {
        for (int i = 0; i < size; i++) {
            nearbySide(tile.x, tile.y, rotation, i, Tmp.p1);
            for (int j = 0; j < range; j++) {
                Tile other = world.tile(Tmp.p1.x + Geometry.d4x(rotation) * j, Tmp.p1.y + Geometry.d4y(rotation) * j);
                if (other != null && other.solid()) {
                    Item drop = other.wallDrop();
                    if (drop != null && drop.hardness <= tier) {
                        return true;
                    }
                    break;
                }
            }
        }
        return false;
    }

    public class GasBeamDrillBuild extends GasBuilding {

        public Tile[] facing = new Tile[size];

        public Point2[] lasers = new Point2[size];

        @Nullable
        public Item lastItem;

        public float time;

        public float warmup, boostWarmup;

        public float lastDrillSpeed;

        @Override
        public void drawSelect() {
            if (lastItem != null) {
                float dx = x - size * tilesize / 2f, dy = y + size * tilesize / 2f, s = iconSmall / 4f;
                Draw.mixcol(Color.darkGray, 1f);
                Draw.rect(lastItem.fullIcon, dx, dy - 1, s, s);
                Draw.reset();
                Draw.rect(lastItem.fullIcon, dx, dy, s, s);
            }
        }

        @Override
        public void updateTile() {
            super.updateTile();
            if (lasers[0] == null)
                updateLasers();
            warmup = Mathf.approachDelta(warmup, Mathf.num(efficiency > 0), 1f / 60f);
            lastItem = null;
            boolean multiple = false;
            int dx = Geometry.d4x(rotation), dy = Geometry.d4y(rotation), facingAmount = 0;
            // update facing tiles
            for (int p = 0; p < size; p++) {
                Point2 l = lasers[p];
                Tile dest = null;
                for (int i = 0; i < range; i++) {
                    int rx = l.x + dx * i, ry = l.y + dy * i;
                    Tile other = world.tile(rx, ry);
                    if (other != null) {
                        if (other.solid()) {
                            Item drop = other.wallDrop();
                            if (drop != null && drop.hardness <= tier) {
                                facingAmount++;
                                if (lastItem != drop && lastItem != null) {
                                    multiple = true;
                                }
                                lastItem = drop;
                                dest = other;
                            }
                            break;
                        }
                    }
                }
                facing[p] = dest;
            }
            // when multiple items are present, count that as no item
            if (multiple) {
                lastItem = null;
            }
            float multiplier = Mathf.lerp(1f, optionalBoostIntensity, optionalEfficiency);
            boostWarmup = Mathf.lerpDelta(boostWarmup, optionalEfficiency, 0.1f);
            lastDrillSpeed = (facingAmount * multiplier * timeScale) / drillTime;
            time += edelta() * multiplier;
            if (time >= drillTime) {
                for (var tile : facing) {
                    Item drop = tile == null ? null : tile.wallDrop();
                    if (items.total() < itemCapacity && drop != null) {
                        items.add(drop, 1);
                        produced(drop);
                    }
                }
                time %= drillTime;
            }
            if (timer(timerDump, dumpTime)) {
                dump();
            }
        }

        @Override
        public boolean shouldConsume() {
            return items.total() < itemCapacity && enabled;
        }

        @Override
        public void draw() {
            Draw.rect(block.region, x, y);
            Draw.rect(topRegion, x, y, rotdeg());
            if (isPayload())
                return;
            var dir = Geometry.d4(rotation);
            int ddx = Geometry.d4x(rotation + 1), ddy = Geometry.d4y(rotation + 1);
            for (int i = 0; i < size; i++) {
                Tile face = facing[i];
                if (face != null) {
                    Point2 p = lasers[i];
                    float lx = face.worldx() - (dir.x / 2f) * tilesize, ly = face.worldy() - (dir.y / 2f) * tilesize;
                    float width = (laserWidth + Mathf.absin(Time.time + i * 5 + id * 9, glowScl, pulseIntensity)) * warmup;
                    Draw.z(Layer.power - 1);
                    Draw.mixcol(glowColor, Mathf.absin(Time.time + i * 5 + id * 9, glowScl, glowIntensity));
                    if (Math.abs(p.x - face.x) + Math.abs(p.y - face.y) == 0) {
                        Draw.scl(width);
                        if (boostWarmup < 0.99f) {
                            Draw.alpha(1f - boostWarmup);
                            Draw.rect(laserCenter, lx, ly);
                        }
                        if (boostWarmup > 0.01f) {
                            Draw.alpha(boostWarmup);
                            Draw.rect(laserCenterBoost, lx, ly);
                        }
                        Draw.scl();
                    } else {
                        float lsx = (p.x - dir.x / 2f) * tilesize, lsy = (p.y - dir.y / 2f) * tilesize;
                        if (boostWarmup < 0.99f) {
                            Draw.alpha(1f - boostWarmup);
                            Drawf.laser(laser, laserEnd, lsx, lsy, lx, ly, width);
                        }
                        if (boostWarmup > 0.001f) {
                            Draw.alpha(boostWarmup);
                            Drawf.laser(laserBoost, laserEndBoost, lsx, lsy, lx, ly, width);
                        }
                    }
                    Draw.color();
                    Draw.mixcol();
                    Draw.z(Layer.effect);
                    Lines.stroke(warmup);
                    rand.setState(i, id);
                    Color col = face.wallDrop().color;
                    Color spark = Tmp.c3.set(sparkColor).lerp(boostHeatColor, boostWarmup);
                    for (int j = 0; j < sparks; j++) {
                        float fin = (Time.time / sparkLife + rand.random(sparkRecurrence + 1f)) % sparkRecurrence;
                        float or = rand.range(2f);
                        Tmp.v1.set(sparkRange * fin, 0).rotate(rotdeg() + rand.range(sparkSpread));
                        Draw.color(spark, col, fin);
                        float px = Tmp.v1.x, py = Tmp.v1.y;
                        if (fin <= 1f)
                            Lines.lineAngle(lx + px + or * ddx, ly + py + or * ddy, Angles.angle(px, py), Mathf.slope(fin) * sparkSize);
                    }
                    Draw.reset();
                }
            }
            if (glowRegion.found()) {
                Draw.z(Layer.blockAdditive);
                Draw.blend(Blending.additive);
                Draw.color(Tmp.c1.set(heatColor).lerp(boostHeatColor, boostWarmup), warmup * (heatColor.a * (1f - heatPulse + Mathf.absin(heatPulseScl, heatPulse))));
                Draw.rect(glowRegion, x, y, rotdeg());
                Draw.blend();
                Draw.color();
            }
            Draw.blend();
            Draw.reset();
        }

        @Override
        public void onProximityUpdate() {
            // when rotated.
            updateLasers();
        }

        void updateLasers() {
            for (int i = 0; i < size; i++) {
                if (lasers[i] == null)
                    lasers[i] = new Point2();
                nearbySide(tileX(), tileY(), rotation, i, lasers[i]);
            }
        }

        @Override
        public byte version() {
            return 1;
        }

        @Override
        public void write(Writes write) {
            super.write(write);
            write.f(time);
            write.f(warmup);
        }

        @Override
        public void read(Reads read, byte revision) {
            super.read(read, revision);
            if (revision >= 1) {
                time = read.f();
                warmup = read.f();
            }
        }
    }
}
